from pathlib import Path
from typing import Callable, List, Tuple

import h5py  # type: ignore
import numpy as np
import open3d  # type: ignore
import torch
import torch.utils.data as data
from torch import Tensor

from pycarus.geometry.pcd import get_o3d_from_tensor
from pycarus.transforms.var import Compose

T_ITEM = Tuple[str, str, Tensor, Tensor]


class MVP(data.Dataset):
    def __init__(
        self,
        root: Path,
        split: str,
        num_points: int = 2048,
        novel_input: bool = True,
        novel_input_only: bool = False,
        categories: List[str] = [],
        download: bool = False,
        transforms: List[Callable] = [],
    ) -> None:
        """Class implementing the Completion3D dataset as proposed in:

        Pan, L., Chen, X., Cai, Z., Zhang, J., Zhao, H., Yi, S., & Liu, Z. (2021).
        Variational Relational Point Completion Network.
        In Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition (pp. 8524-8533).

        Args:
            root: The path to the folder containing the dataset.
            split: The name of the split to load. Allowed values: train, val, test.
            num_points: The resolution for groundtruth clouds. Allowed values: 2048, 4096, 8192, 16384.
            novel_input: Use the novel shapes introduced in MVP paper. Default True.
            novel_input_only: Use only the novel shapes introduced in MVP paper. Default True.
            categories: A list of categories id to select. Defaults to [].
            download: If True download the dataset using the "tmp" folder. Defaults to False.
            transforms: The transform to apply to the item. Defaults to [].

        Raises:
            FileNotFoundError: If the folder does not exist.
            ValueError: If the chosen split is not allowed.
            ValueError: If the chosen resolution is not allowed.
        """
        splits = ["train", "val", "test"]
        if split not in splits:
            raise ValueError(f"Split {split} not allowed, only {splits} are allowed.")

        val = split == "val"
        self.split = split if split != "val" else "train"

        self.num_points = num_points
        res_gt = [2048, 4096, 8192, 16384]
        if self.num_points not in res_gt:
            raise ValueError(f"{self.num_points} value not allowed, only {res_gt} are allowed.")

        self.root = root
        if not download and not self.root.is_dir():
            raise FileNotFoundError(f"{self.root} not found.")

        if download:
            if self.root.exists():
                print("Dataset root already exists, not downloading.")
            else:
                raise ValueError(
                    "Download not avaible for this dataset."
                    " You can manually download it at "
                    "https://paul007pl.github.io/projects/VRCNet.html"
                )

        self.transforms = Compose(transforms)

        self.path_inc = self.root / f"mvp_{self.split}_input.h5"
        file_inc = h5py.File(self.path_inc, "r")
        self.incompletes = np.array(file_inc.get("incomplete_pcds"))
        self.labels = np.array(file_inc.get("labels"))
        incompletes_novel = np.array(file_inc.get("novel_incomplete_pcds"))
        labels_novel = np.array(file_inc.get("novel_labels"))
        file_inc.close()

        self.path_comp = self.root / f"mvp_{self.split}_gt_{self.num_points}pts.h5"
        file_comp = h5py.File(self.path_comp, "r")
        self.completes = np.array(file_comp.get("complete_pcds"))
        completes_novel = np.array(file_comp.get("novel_complete_pcds"))
        file_comp.close()

        if novel_input_only:
            self.incompletes = incompletes_novel
            self.completes = completes_novel
            self.labels = labels_novel
        elif novel_input:
            self.incompletes = np.concatenate((self.incompletes, incompletes_novel), axis=0)  # type: ignore
            self.completes = np.concatenate((self.completes, completes_novel), axis=0)  # type: ignore
            self.labels = np.concatenate((self.labels, labels_novel), axis=0)  # type: ignore

        self.indices = list(range(self.incompletes.shape[0]))
        if self.split != "test":
            val_indices = list(range(0, len(self.indices), 10))
            if val:
                self.indices = val_indices
            else:
                self.indices = [i for i in self.indices if i not in val_indices]

        self.all_categories = [
            "02691156",
            "02933112",
            "02958343",
            "03001627",
            "03636649",
            "04256520",
            "04379243",
            "04530566",
            "02818832",
            "02828884",
            "02871439",
            "02924116",
            "03467517",
            "03790512",
            "03948459",
            "04225987",
        ]
        self.categories = categories if categories else self.all_categories
        self.indices = self.filter_indices()

    def filter_indices(self) -> List[int]:
        filtered_indices = []
        for index in self.indices:
            label = int(self.labels[index])
            if self.all_categories[label] in self.categories:
                filtered_indices.append(index)
        return filtered_indices

    def __len__(self) -> int:
        return len(self.indices)

    def __getitem__(self, index: int) -> T_ITEM:
        idx = self.indices[index]
        idx_complete = idx // 26
        idx_view = idx - idx_complete
        incomplete = self.transforms(torch.from_numpy((self.incompletes[idx])))
        complete = self.transforms(torch.from_numpy((self.completes[idx_complete])))
        label = torch.tensor(self.labels[idx], dtype=torch.long)
        category = self.all_categories[label]
        name = f"s{idx_complete}_v{idx_view:02d}"

        return category, name, incomplete, complete

    @classmethod
    def show_item(cls, sample: T_ITEM) -> open3d.geometry.PointCloud:
        """Prepare one item in order to be visualized using open3D draw geometries.
        Args:
            sample: The sample to show.
        Returns:
            The point cloud to draw with open 3D.
        """
        _, _, partial, complete = sample
        pcd_partial = get_o3d_from_tensor(partial)

        complete = complete + torch.tensor([1.0, 0.0, 0.0])
        pcd_complete = get_o3d_from_tensor(complete)

        return pcd_complete + pcd_partial

    def get_categories(self) -> List[str]:
        """Get the list of loaded categories.

        Returns:
            A list containing the ids of the loaded categories.
        """
        return self.categories



from pathlib import Path

import open3d as o3d  # type: ignore

from pycarus.datasets.shapenet_part import ShapeNetPartSegmentation


def main() -> None:
    path_ds = "../ShapeNetPart/"
    dataset = ShapeNetPartSegmentation(path_root=Path(path_ds), split="train", download=True)
    print(len(dataset))

    for index, sample in enumerate(dataset):  # type: ignore
        name = dataset.samples[index]
        category = dataset.ids_to_categories[dataset.classes[sample[1]]]

        print(f"Name: {name[0]} - Category: {category}.")
        o3d.visualization.draw_geometries([ShapeNetPartSegmentation.show_item(sample)])


if __name__ == "__main__":
    main()



from collections import defaultdict
from pathlib import Path
from statistics import mean
from typing import cast

import torch
from torch import Tensor

from pycarus.geometry.pcd import read_pcd
from pycarus.metrics.chamfer_distance import chamfer
from pycarus.metrics.emd_distance import emd
from pycarus.metrics.f_score import fscore
from pycarus.utils import progress_bar


def pcd_completion_evaluation(
    gt_dir: Path,
    pred_dir: Path,
    results_dir: Path,
    batch_size: int = 512,
    gpu: bool = True,
) -> None:
    """Evaluate point cloud completion results.

    The function requires the ground-truth clouds and the predicted ones to be
    organized in two separate directories with the same structure. Each directory
    must contain one sub-directory for each category and each subdirectory must
    contain one .ply file for each cloud. The function expects the name of the files
    in the "gt_dir" and "pred_dir" directories to be matching in order to perform
    the comparison.

    Four metrics are computed: the two Chamfer Distances proposed in the paper
    "Variational Relational Point Completion Network.", Earth Mover's Distance and F1 score.
    Results are saved into two files ("global_results.csv" and "single_results.csv")
    inside the given "result_dir" directory.

    Args:
        gt_dir: Directory with ground-truth clouds.
        pred_dir: Directory with predicted clouds.
        results_dir: Directory where the .csv files with the results will be saved.
        batch_size: The batch size to use for computing metrics.
        gpu: Whether to use gpu or not to compute metrics.

    Raises:
        ValueError: If one prediction is missing.
    """
    results_dir.mkdir(parents=True, exist_ok=True)

    single_results_file = results_dir / "single_results.csv"
    if single_results_file.exists():
        single_results_file.unlink()

    with open(single_results_file, "wt") as f:
        f.write("CATEGORY/NAME,CD_P,CD_T,EMD,F1-SCORE\n")

    cd_ps = defaultdict(list)
    cd_ts = defaultdict(list)
    emds = defaultdict(list)
    f1_scores = defaultdict(list)

    gt_files = []
    pred_files = []
    categories_dirs = [subdir for subdir in gt_dir.iterdir() if subdir.is_dir()]
    for category_dir in categories_dirs:
        for gt_file in category_dir.glob("*.ply"):
            gt_files.append(gt_file)
            pred_files.append(pred_dir / category_dir.name / gt_file.name)

    num_batches = len(gt_files) // batch_size
    num_batches = num_batches + 1 if len(gt_files) % batch_size != 0 else num_batches

    for batch in progress_bar(range(num_batches)):
        start = batch * batch_size
        end = start + batch_size
        gt_files_batch = gt_files[start:end]
        pred_files_batch = pred_files[start:end]

        gts = torch.stack([read_pcd(gt_file) for gt_file in gt_files_batch], dim=0)
        preds = torch.stack([read_pcd(pred_file) for pred_file in pred_files_batch], dim=0)

        if gpu:
            gts = gts.cuda()
            preds = preds.cuda()

        cds_pred_gt, cds_gt_pred = chamfer(preds, gts, squared=False)
        cds_pred_gt_sq, cds_gt_pred_sq = chamfer(preds, gts, squared=True)
        emds_, _ = emd(preds, gts, eps=0.004, iterations=3000, squared=False)
        f1s, _, _ = fscore(preds.cpu(), gts.cpu(), 0.01)

        for i, gt_file in enumerate(gt_files_batch):
            category = gt_file.parent.name
            shape_name = gt_file.stem

            cd_p = float(((cds_gt_pred[i] + cds_pred_gt[i]) / 2).item())
            cd_ps["all"].append(cd_p)
            cd_ps[category].append(cd_p)

            cd_t = float((cds_gt_pred_sq[i] + cds_pred_gt_sq[i]).item())
            cd_ts["all"].append(cd_t)
            cd_ts[category].append(cd_t)

            emd_ = float(emds_[i].item())
            emds["all"].append(emd_)
            emds[category].append(emd_)

            f1_score = float(cast(Tensor, f1s)[i].item())
            f1_scores["all"].append(f1_score)
            f1_scores[category].append(f1_score)

            with open(single_results_file, "at") as f:
                line = f"{category}/{shape_name},{cd_p:.8f},{cd_t:.8f},{emd_:.8f},{f1_score:.8f}\n"
                f.write(line)

    global_results_file = results_dir / "global_results.csv"
    if global_results_file.exists():
        global_results_file.unlink()

    with open(global_results_file, "wt") as f:
        f.write("CATEGORY,CD_P,CD_T,EMD,F1-SCORE\n")

    for category in sorted(cd_ps.keys()):
        if category != "all":
            m_cd_p = mean(cd_ps[category])
            m_cd_t = mean(cd_ts[category])
            m_emd_ = mean(emds[category])
            m_f1_score = mean(f1_scores[category])

            with open(global_results_file, "at") as f:
                f.write(f"{category},{m_cd_p:.8f},{m_cd_t:.8f},{m_emd_:.8f},{m_f1_score:.8f}\n")

    with open(global_results_file, "at") as f:
        f.write("\n\n")
        f.write(f"MEAN CD_P,{mean(cd_ps['all']):.8f}\n")
        f.write(f"MEAN CD_T,{mean(cd_ts['all']):.8f}\n")
        f.write(f"MEAN EMD,{mean(emds['all']):.8f}\n")
        f.write(f"MEAN F1 SCORE,{mean(f1_scores['all']):.8f}\n")



from typing import Tuple

import numpy as np
import torch
from skimage.measure import marching_cubes  # type: ignore
from torch import nn


@torch.no_grad()  # type: ignore
def sdf_to_mesh(
    mlp: nn.Module,
    grid_size: int = 256,
    max_batch: int = 100_000,
) -> Tuple[np.ndarray, np.ndarray]:
    """Obtain the mesh corresponding to the 3D shape whose SDF
    is contained in the input MLP.

    Args:
        mlp: The MLP representing the 3D shape SDF.
        grid_size: The SDF sampling resolution. Defaults to 256.
        max_batch: The max batch size used to query the input MLP. Defaults to 100_000.

    Returns:
        - The vertices of the computed mesh as numpy array.
        - The faces of the computed mesh as numpy array.
    """
    voxel_origin = [-1, -1, -1]
    voxel_size = 2.0 / (grid_size - 1)

    overall_index = torch.arange(0, grid_size ** 3, 1)
    coords = torch.zeros(grid_size ** 3, 4)

    coords[:, 2] = overall_index % grid_size
    coords[:, 1] = (overall_index.long() / grid_size) % grid_size
    coords[:, 0] = ((overall_index.long() / grid_size) / grid_size) % grid_size

    coords[:, 0] = (coords[:, 0] * voxel_size) + voxel_origin[2]
    coords[:, 1] = (coords[:, 1] * voxel_size) + voxel_origin[1]
    coords[:, 2] = (coords[:, 2] * voxel_size) + voxel_origin[0]

    num_samples = grid_size ** 3
    start = 0

    while start < num_samples:
        end = min(start + max_batch, num_samples)
        coords_subset = coords[start:end, :3].cuda()
        sdf = mlp(coords_subset).cpu()
        coords[start:end, 3] = sdf
        start += max_batch

    sdf = coords[:, 3].reshape(grid_size, grid_size, grid_size)

    verts, faces, _, _ = marching_cubes(sdf.numpy(), level=0.0, spacing=[voxel_size] * 3)

    return verts, faces



import shutil
from pathlib import Path
from typing import Callable, List, Tuple

import h5py  # type: ignore
import numpy as np
import open3d  # type: ignore
import torch
from torch.utils.data import Dataset

from pycarus.geometry.pcd import get_o3d_from_tensor
from pycarus.transforms.var import Compose
from pycarus.utils import download_and_extract

T_ITEM = Tuple[str, str, torch.Tensor, torch.Tensor]


class Completion3D(Dataset):
    def __init__(
        self,
        root: Path,
        split: str,
        categories: List[str] = [],
        download: bool = False,
        transforms: List[Callable] = [],
    ) -> None:
        """Class implementing the Completion3D dataset as proposed in:

        Tchapmi, L. P., Kosaraju, V., Rezatofighi, H., Reid, I., & Savarese, S. (2019).
        Topnet: Structural point cloud decoder.
        In Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition (pp. 383-392).

        Args:
            root: The path to the folder containing the dataset.
            split: The name of the split to load.
            categories: A list of categories id to select. Defaults to [].
            download: If True download the dataset using the "tmp" folder. Defaults to False.
            transforms: The transform to apply to the item. Defaults to [].

        Raises:
            FileNotFoundError: If the folder does not exist.
            ValueError: If the chosen split is not allowed.
        """
        super().__init__()

        self.split = split
        self.splits = ["train", "val", "test"]
        if self.split not in self.splits:
            raise ValueError(f"{self.split} value not allowed, only allowed {self.splits}.")

        self.root = root
        if not download and not self.root.is_dir():
            raise FileNotFoundError(f"{self.root} not found.")

        self.transform = Compose(transforms)

        if download:
            if self.root.exists():
                print("Dataset root already exists, not downloading.")
            else:
                self.url = "http://download.cs.stanford.edu/downloads/completion3d/dataset2019.zip"
                self.__download()

        self.list_file, cat_in_file = self.read_samples(
            self.root / f"{self.split}.list", categories
        )
        self.categories = categories if categories else cat_in_file

    def __getitem__(self, index: int) -> T_ITEM:
        sample = self.list_file[index]
        id_category, name = sample.split("/")[0], sample.split("/")[1]

        path_partial = self.root / self.split / "partial" / id_category / f"{name}.h5"
        partial = torch.tensor(np.array(h5py.File(path_partial, "r")["data"]), dtype=torch.float)

        if self.split == "test":
            complete = torch.zeros_like(partial)
        else:
            path_complete = self.root / self.split / "gt" / id_category / f"{name}.h5"
            complete = torch.tensor(
                np.array(h5py.File(path_complete, "r")["data"]), dtype=torch.float
            )

        return id_category, name, partial, complete

    def __len__(self) -> int:
        return len(self.list_file)

    @staticmethod
    def read_samples(path: Path, filter_categories: List[str]) -> Tuple[List[str], List[str]]:
        """Read a text file.

        Args:
            path: the path to the file to read.

        Returns:
            A tuple containing:
            - The list with one line of the file as sample.
            - The list of categories.
        """
        list_files: List[str] = []
        list_categories: List[str] = []
        with open(path, "rt") as f:
            for line in f.readlines():
                name_sample = line.rstrip()
                category = name_sample.split("/")[0]
                if category not in list_categories:
                    list_categories.append(category)

                if filter_categories:
                    if category in filter_categories:
                        list_files.append(name_sample)
                else:
                    list_files.append(name_sample)

        return list_files, list_categories

    def __download(self) -> None:
        """Function to download the dataset."""
        path_temp = Path("/tmp")
        path_file_downloaded = path_temp / Path(self.url.split("/")[-1])
        download_and_extract(self.url, path_file_downloaded, path_temp, None)

        path_ds_extracted = path_temp / "shapenet"

        shutil.copytree(str(path_ds_extracted), self.root)
        shutil.rmtree(str(path_ds_extracted))

    @classmethod
    def show_item(cls, sample: T_ITEM) -> open3d.geometry.PointCloud:
        """Prepare one item in order to be visualized using open3D draw geometries.
        Args:
            sample: The sample to show.
        Returns:
            The point cloud to draw with open 3D.
        """
        _, _, partial, complete = sample
        pcd_partial = get_o3d_from_tensor(partial)

        complete = complete + torch.tensor([1.0, 0.0, 0.0])
        pcd_complete = get_o3d_from_tensor(complete)

        return pcd_complete + pcd_partial

    def get_categories(self) -> List[str]:
        """Get the list of loaded categories.

        Returns:
            A list containing the ids of the loaded categories.
        """
        return self.categories



import pytorch3d.transforms.transform3d as pt3d  # type: ignore
import torch
from pytorch3d.transforms.rotation_conversions import _axis_angle_rotation  # type: ignore
from torch import Tensor

from pycarus.geometry.pcd import apply_affine_to_pcd, farthest_point_sampling, jitter_pcd
from pycarus.geometry.pcd import normalize_pcd, random_drop_points, random_point_sampling
from pycarus.geometry.pcd import rotate_pcd, scale_pcd, shuffle_pcd, translate_pcd


class SamplePcd:
    def __init__(self, num_points: int, algorithm: str) -> None:
        """Subsample the input point cloud(s) using the given algorithm.

        Args:
            num_points: Number of point to sample.
            algorithm: the algorithm to use for the subsampling between ["fps", "random"].
        """
        self.num_points = num_points
        self.algorithm = algorithm

        if self.algorithm not in ["fps", "random"]:
            raise ValueError(f"Expected type to be one of [fps, random], got {algorithm}")

    def __call__(self, pcd: Tensor) -> Tensor:
        """Sample point cloud.

        Args:
            pcd: The input point cloud with shape (NUM_POINTS, D).

        Returns:
            The sampled point cloud.
        """
        pcd_ss = torch.tensor([])
        if self.algorithm == "fps":
            pcd_ss = farthest_point_sampling(pcd, self.num_points)

        if self.algorithm == "random":
            pcd_ss = random_point_sampling(pcd, self.num_points)

        return pcd_ss

    def __repr__(self) -> str:
        format_string = f"{self.__class__.__name__}: num_points: {self.num_points}"
        format_string += f" - algorithm: {self.algorithm}."
        return format_string


class NormalizePcd:
    """Normalize a given point cloud into the unit cube."""

    def __call__(self, pcd: Tensor) -> Tensor:
        """Normalize a point cloud.

        Args:
            pcd: The input point cloud with shape (NUM_POINTS, D).

        Returns:
            The normalized point cloud.
        """
        return normalize_pcd(pcd)


class AffinePcd(pt3d.Transform3d):
    def __init__(self, matrix: Tensor) -> None:
        """Transform to apply an affine transformation, i.e. rotation plus a translation.

        Args:
            matrix: The motion matrix with shape (4, 4).
        """
        self.matrix = matrix
        super().__init__(matrix=self.matrix)

    def __call__(self, pcd: Tensor) -> Tensor:
        return apply_affine_to_pcd(pcd, self.matrix)


class RotatePcd(pt3d.Rotate):
    def __init__(self, rotation: Tensor) -> None:
        """Transform to rotate a point cloud.

        Args:
            rotation: The rotation matrix with shape (3, 3).
        """
        self.rotation = rotation
        super().__init__(self.rotation)

    def __call__(self, pcd: Tensor) -> Tensor:
        return rotate_pcd(pcd, self.rotation)


class RandomRotatePcdAroundAxis:
    def __init__(self, axis: str, min_angle: float = 0, max_angle: float = 180):
        """Rotate a point cloud around one axis by a random angle sampled within a range.

        Args:
            min_angle: the lower bound for the angle in radians.
            max_angle: the upper bound for the angle in radians.
            axis: the axis of rotation.

        Raises:
            ValueError: if the required axis is different from [X, Y, Z].
        """
        self.axis = axis.upper()

        if axis not in ["X", "Y", "Z"]:
            raise ValueError(f"Expected axis to be one of [X, Y, Z], got {axis}")

        self.min_angle = float(torch.deg2rad(torch.tensor(float(min_angle))).item())
        self.max_angle = float(torch.deg2rad(torch.tensor(float(max_angle))).item())
        self.last_matrix = torch.eye(3)

    @staticmethod
    def _get_params(min_angle: float, max_angle: float, axis: str) -> Tensor:
        """Get random rotation matrix with an angle in the range min_angle - max_angle.

        Args:
            min_angle: the lower bound for the angle in radians.
            max_angle: the upper bound for the angle in radians.
            axis: the axis of rotation.

        Returns:
            The rotation matrix with shape (3, 3).
        """
        theta = torch.FloatTensor(1).uniform_(min_angle, max_angle)
        matrix_col_major = _axis_angle_rotation(axis, theta)
        # The convention used in this framework is to operate on points as row vectors.
        # The rotation matrix returned from _axis_angle_rotation is for transforming column vectors.
        # Therefore we transpose this matrix.
        matrix_row_major = matrix_col_major.transpose(1, 2)[0]
        return matrix_row_major

    def __call__(self, pcd: Tensor) -> Tensor:
        self.last_matrix = self._get_params(self.min_angle, self.max_angle, self.axis)
        return rotate_pcd(pcd, self.last_matrix)

    def __repr__(self) -> str:
        format_string = f"{self.__class__.__name__}: min_angle: {self.min_angle}"
        format_string += f" - max_angle: {self.max_angle} - axis: {self.axis}."

        return format_string


class TranslatePcd(pt3d.Translate):
    def __init__(self, translation: Tensor) -> None:
        """Transform to translate a point cloud.

        Args:
            translation: The translation offset with shape (3,).
        """
        self.translation = translation
        x, y, z = translation[0], translation[1], translation[2]
        super().__init__(x=x, y=y, z=z)

    def __call__(self, pcd: Tensor) -> Tensor:
        return translate_pcd(pcd, self.translation)


class RandomTranslatePcd:
    def __init__(self, min_tr: float = -0.5, max_tr: float = 0.5) -> None:
        """Transform to translate randomly a point cloud.

        Args:
            min_tr: The minimum translation.
            max_tr: The maximum translation.
        """
        self.min_tr = min_tr
        self.max_tr = max_tr

    def __call__(self, pcd: Tensor) -> Tensor:
        transl = torch.rand((3,)) * (self.max_tr - self.min_tr) + self.min_tr
        return translate_pcd(pcd, transl)


class ScalePcd(pt3d.Scale):
    def __init__(self, scale_factor: Tensor) -> None:
        """Transform to scale a point cloud.

        Args:
            scale_factor: The scale factor for the x, y, z, dimensions with shape (3,).
        """
        self.scale_factor = scale_factor
        x, y, z = scale_factor[0], scale_factor[1], scale_factor[2]
        super().__init__(x=x, y=y, z=z)

    def __call__(self, pcd: Tensor) -> Tensor:
        return scale_pcd(pcd, self.scale_factor)


class RandomScalePcd:
    def __init__(self, min_scale: float = 0.5, max_scale: float = 1.5) -> None:
        """Transform to scale randomly a point cloud.

        Args:
            min_scale: The minimum scale factor.
            max_scale: The maximum scale factor.
        """
        self.min_scale = min_scale
        self.max_scale = max_scale

    def __call__(self, pcd: Tensor) -> Tensor:
        scale = torch.rand((3,)) * (self.max_scale - self.min_scale) + self.min_scale
        return scale_pcd(pcd, scale)


class JitterPcd:
    def __init__(self, sigma: float, clip: float) -> None:
        """Jitter point cloud by adding random noise.

        Args:
            sigma: The sigma for the gaussian noise.
            clip: The clipping value.
        """
        self.sigma = sigma
        self.clip = clip

    def __call__(self, pcd: Tensor) -> Tensor:
        return jitter_pcd(pcd, self.sigma, self.clip)


class RandomDropPcd:
    def __init__(self, min_percentage: float, max_percentage: float) -> None:
        """Remove random percentage of points from a point cloud. The number of points is uniformly
        sampled in the range min_percentage - max_percentage.

        Args:
            min_percentage: The lower bound for the percentage of points to remove.
            max_percentage: The upper bound for the percentage of points to remove.
        """
        self.min_percentage = min_percentage
        self.max_percentage = max_percentage
        self.drop_percentage = 0.0
        self.last_kept = torch.tensor([])
        self.last_removed = torch.tensor([])

    @staticmethod
    def _get_params(min_percentage: float, max_percentage: float) -> float:
        """Get random number in the range min_percentage - max_percentage.

        Args:
            min_percentage: the lower bound for the percentage.
            max_percentage: the upper bound for the percentage.

        Returns:
            A random number in the specified range.
        """
        random_percentage = torch.FloatTensor(1).uniform_(min_percentage, max_percentage)[0].item()
        return float(random_percentage)

    def __call__(self, pcd: Tensor) -> Tensor:
        self.last_drop_percentage = self._get_params(self.min_percentage, self.max_percentage)
        points, indices_kept, indices_removed = random_drop_points(pcd, self.last_drop_percentage)
        self.last_kept = indices_kept
        self.last_removed = indices_removed

        return points

    def __repr__(self) -> str:
        format_string = f"{self.__class__.__name__}: min_percentage: {self.min_percentage}"
        format_string += f" - max_percentage: {self.max_percentage}."

        return format_string


class ShufflePcd:
    """Shuffle the order of the points inside the point cloud."""

    def __call__(self, pcd: Tensor) -> Tensor:
        return shuffle_pcd(pcd)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"



from typing import Callable, List

from torch import Tensor


class Compose:
    def __init__(self, transforms: List[Callable]) -> None:
        """Compose several transforms together.

        Each transform should be an instance of a class with a __call__ method.

        Args:
            transforms: the transforms to apply.
        """
        self.transforms = transforms

    def __call__(self, data: Tensor) -> Tensor:
        """Apply the transforms to the input data.

        Args:
            data: The data to transform.

        Returns:
            The transformed data.
        """
        for t in self.transforms:
            data = t(data)
        return data

    def __repr__(self) -> str:
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
            format_string += "\n)"
        return format_string



import os
from importlib.util import find_spec
from typing import Tuple

import torch

from pycarus.geometry.pcd import batchify, unbatchify


def chamfer(
    prediction: torch.Tensor, groundtruth: torch.Tensor, squared: bool = True
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Compute the Chamfer Distance defined as in:

    Groueix, T., Fisher, M., Kim, V. G., Russell, B. C., & Aubry, M. (2018).
    A papier-mâché approach to learning 3d surface generation.
    In Proceedings of the IEEE conference on computer vision and pattern recognition (pp. 216-224).

    Args:
        prediction: The source point cloud(s) with shape ([B,] NUM_POINTS, 3).
        groundtruth: The target point cloud(s) with shape ([B,] NUM_POINTS, 3).
        squared: If true return the squared euclidean distance.

    Raises:
        ModuleNotFoundError: If the cuda extension is not installed.

    Returns:
        A tuple containing:
        - A tensor with the average distance from the prediction to the groundtruth.
        - A tensor with the average distance from the groundtruth to the prediction.
    """
    pcd_src = groundtruth
    pcd_trg = prediction

    use_cuda = pcd_src.device != torch.device("cpu")

    batched, [pcd_src, pcd_trg] = batchify([pcd_src, pcd_trg], 3)

    if use_cuda:
        not_chamfer_cuda = find_spec("chamfer_cuda") is None

        if not_chamfer_cuda:
            folder = os.path.dirname(os.path.realpath(__file__))
            str_exp = f"Chamfer Cuda module not found. Install it running {folder}/external/setup_chamfer_gpu.py"
            raise ModuleNotFoundError(str_exp)
        else:
            from pycarus.metrics.external.chamfer_gpu import Chamfer_GPU  # type: ignore

            chamfer = Chamfer_GPU()
            dist_x_y, dist_y_x, _, _ = chamfer(pcd_src, pcd_trg)
    else:
        x, y = pcd_src.double(), pcd_trg.double()
        size_batch, num_points_x, _ = x.size()
        size_batch, num_points_y, _ = y.size()

        xx = torch.pow(x, 2).sum(2)
        yy = torch.pow(y, 2).sum(2)
        zz = torch.bmm(x, y.transpose(2, 1))
        rx = xx.unsqueeze(1).expand(size_batch, num_points_y, num_points_x)  # Diagonal elements xx
        ry = yy.unsqueeze(1).expand(size_batch, num_points_x, num_points_y)  # Diagonal elements yy
        corr = rx.transpose(2, 1) + ry - 2 * zz

        dist_x_y = torch.min(corr, 2)[0].float()
        dist_y_x = torch.min(corr, 1)[0].float()

    if not squared:
        dist_x_y, dist_y_x = torch.sqrt(dist_x_y), torch.sqrt(dist_y_x)

    mean_dist_x_y, mean_dist_y_x = dist_x_y.mean(1), dist_y_x.mean(1)
    if batched:
        mean_dist_x_y, mean_dist_y_x = unbatchify([mean_dist_x_y, mean_dist_y_x])

    return mean_dist_x_y, mean_dist_y_x



import json
import shutil
from pathlib import Path
from typing import Callable, Dict, List, Tuple

import numpy as np
import open3d  # type: ignore
import torch
from torch import Tensor
from torch.utils.data import Dataset

from pycarus.geometry.pcd import color_pcd
from pycarus.transforms.var import Compose
from pycarus.utils import download_and_extract

T_ITEM = Tuple[Tensor, int, Tensor]
T_SAMPLES = List[Tuple[str, int]]


class ShapeNetPartSegmentation(Dataset):
    """Class representing the ShapeNet part segmentation dataset as presented in:

    Yi, Li, et al. "A scalable active framework for region annotation in 3d shape collections."
    ACM Transactions on Graphics (ToG) 35.6 (2016): 1-12.
    """

    url = "https://shapenet.cs.stanford.edu/media/shapenetcore_partanno_segmentation_benchmark_v0_normal.zip"
    md5_hash = "8c2f54f22c3f1d0e69da83b86f17f5ec"

    seg_part_to_label = {
        "02691156": [0, 1, 2, 3],
        "02773838": [4, 5],
        "02954340": [6, 7],
        "02958343": [8, 9, 10, 11],
        "03001627": [12, 13, 14, 15],
        "03261776": [16, 17, 18],
        "03467517": [19, 20, 21],
        "03624134": [22, 23],
        "03636649": [24, 25, 26, 27],
        "03642806": [28, 29],
        "03790512": [30, 31, 32, 33, 34, 35],
        "03797390": [36, 37],
        "03948459": [38, 39, 40],
        "04099429": [41, 42, 43],
        "04225987": [44, 45, 46],
        "04379243": [47, 48, 49],
    }

    seg_label_to_color = torch.tensor(
        [
            [0.65, 0.95, 0.05],
            [0.35, 0.05, 0.35],
            [0.65, 0.35, 0.65],
            [0.95, 0.95, 0.65],
            [0.95, 0.65, 0.05],
            [0.35, 0.05, 0.05],
            [0.65, 0.05, 0.05],
            [0.65, 0.35, 0.95],
            [0.05, 0.05, 0.65],
            [0.65, 0.05, 0.35],
            [0.05, 0.35, 0.35],
            [0.65, 0.65, 0.35],
            [0.35, 0.95, 0.05],
            [0.05, 0.35, 0.65],
            [0.95, 0.95, 0.35],
            [0.65, 0.65, 0.65],
            [0.95, 0.95, 0.05],
            [0.65, 0.35, 0.05],
            [0.35, 0.65, 0.05],
            [0.95, 0.65, 0.95],
            [0.95, 0.35, 0.65],
            [0.05, 0.65, 0.95],
            [0.65, 0.95, 0.65],
            [0.95, 0.35, 0.95],
            [0.05, 0.05, 0.95],
            [0.65, 0.05, 0.95],
            [0.65, 0.05, 0.65],
            [0.35, 0.35, 0.95],
            [0.95, 0.95, 0.95],
            [0.05, 0.05, 0.05],
            [0.05, 0.35, 0.95],
            [0.65, 0.95, 0.95],
            [0.95, 0.05, 0.05],
            [0.35, 0.95, 0.35],
            [0.05, 0.35, 0.05],
            [0.05, 0.65, 0.35],
            [0.05, 0.95, 0.05],
            [0.95, 0.65, 0.65],
            [0.35, 0.95, 0.95],
            [0.05, 0.95, 0.35],
            [0.95, 0.35, 0.05],
            [0.65, 0.35, 0.35],
            [0.35, 0.95, 0.65],
            [0.35, 0.35, 0.65],
            [0.65, 0.95, 0.35],
            [0.05, 0.95, 0.65],
            [0.65, 0.65, 0.95],
            [0.35, 0.05, 0.95],
            [0.35, 0.65, 0.95],
            [0.35, 0.05, 0.65],
        ]
    )

    def __init__(
        self,
        path_root: Path,
        split: str,
        download: bool = False,
        transforms: List[Callable] = [],
    ) -> None:
        """Create ShapeNetPartSegmentation dataset.

        Args:
            path_root: The path to the folder containing the dataset.
            split: The name of the split to load.
            download: If True download the dataset using the "tmp" folder. Defaults to False.
            transforms: The transform to apply to the item. Defaults to [].

        Raises:
            ValueError in the split is not in: train, val or test.
        """
        super().__init__()

        splits = ["train", "val", "test"]
        if split not in splits:
            raise ValueError((f"Split {split} found, but expected either " "train, val, or test"))

        self.path_root = path_root
        self.samples: List = []
        self.split = split
        self.transform = Compose(transforms)

        if download:
            if self.path_root.exists():
                print("Dataset root already exists, not downloading.")
            else:
                self.download()

        self.path_file_split = (
            self.path_root
            / Path("train_test_split")
            / Path(f"shuffled_{self.split}_file_list.json")
        )
        classes, class_to_idx = self._get_classes(self.path_root)

        self.samples = self._get_samples(self.path_file_split, class_to_idx)
        self.targets = [s[1] for s in self.samples]
        self.classes = classes
        self.class_to_idx = class_to_idx
        self.ids_to_categories: Dict[str, str] = self._read_ids_to_categories_dict()

    def _read_ids_to_categories_dict(self) -> Dict[str, str]:
        """Read the synsetoffset2category file of ShapeNet dataset.

        Returns:
            A dictionary where the keys are the ids of the categories while the values the name.
        """
        result: Dict[str, str] = {}
        path_file_category = self.path_root / Path("synsetoffset2category.txt")
        with open(path_file_category) as reader:
            content = reader.readlines()
            for element in content:
                element_splitted = element.rstrip().split("\t")
                result[element_splitted[1]] = element_splitted[0]

        return result

    def _get_classes(self, path: Path) -> Tuple[List[str], Dict[str, int]]:
        """Get the classes for the dataset using the folders.

        Args:
            path: the path to the dataset.

        Returns:
            - The classes.
            - The class to index dictionary.
        """
        classes = [
            folder.name
            for folder in path.iterdir()
            if folder.is_dir() and folder.name != "train_test_split"
        ]
        classes.sort()
        class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}

        return classes, class_to_idx

    def _get_samples(self, path_file_split: Path, class_to_idx: Dict[str, int]) -> T_SAMPLES:
        """Get all the samples of the dataset.

        Args:
            path_file_split: path to the split file.
            class_to_idx: the class to label index dictionary.

        Returns:
            The name of the samples together with the class labels.
        """
        file_split = open(str(path_file_split), "r")
        list_samples = json.loads(file_split.read())
        samples: List[Tuple[str, int]] = []

        for sample in list_samples:
            sample_splits = sample.split("/")
            name_class = sample_splits[-2]
            name_sample = sample_splits[-1]
            label = class_to_idx[name_class]

            samples.append((name_sample, label))

        return samples

    def __len__(self) -> int:
        """Get the number of samples in the dataset.

        Returns:
            The number of samples in the dataset.
        """
        return len(self.samples)

    def __getitem__(self, index: int) -> T_ITEM:
        """Get one sample.

        Args:
            index: The index of the requested element.

        Returns:
            - the point cloud as a torch tensor with shape (NUM_POINTS, D).
            - the class label.
            - the segmentation part labels with shape (NUM_POINTS, D).
        """
        name_sample, label = self.samples[index]
        name_category = self.classes[label]
        path_file_sample = self.path_root / Path(name_category, f"{name_sample}.txt")
        pcd_norm_label = np.array(np.loadtxt(str(path_file_sample)))  # type: ignore
        pcd = torch.tensor(pcd_norm_label[:, 0:3], dtype=torch.float)
        parts_labels = torch.tensor(pcd_norm_label[:, -1], dtype=torch.long)
        parts_labels = parts_labels.view(-1, 1)

        pcd_and_labels = torch.cat((pcd, parts_labels), dim=1)
        pcd_and_labels = self.transform(pcd_and_labels)
        pcd = pcd_and_labels[:, :3]
        parts_labels = pcd_and_labels[:, 3:].long().view(-1)

        return pcd, label, parts_labels

    def download(self) -> None:
        """Function to download the dataset."""
        path_temp = Path("/tmp")
        path_file_downloaded = path_temp / Path(self.url.split("/")[-1])
        download_and_extract(self.url, path_file_downloaded, path_temp, self.md5_hash)

        path_ds_extracted = path_temp / Path(
            "shapenetcore_partanno_segmentation_benchmark_v0_normal"
        )

        shutil.copytree(str(path_ds_extracted), self.path_root)
        shutil.rmtree(str(path_ds_extracted))

    @classmethod
    def show_item(cls, sample: T_ITEM) -> open3d.geometry.PointCloud:
        """Prepare one item in order to be visualized using open3D draw geometries.

        Args:
            sample: The sample to show.

        Returns:
            The point cloud to draw with open 3D.
        """
        pcd, _, parts_labels = sample
        col = cls.seg_label_to_color[parts_labels, :]

        return color_pcd(pcd, col)
